#include "MCP2515.h"
#include "SPI.h"

extern mcp2515_RESET()
{
 
  SetSS(0);				   //select slave
  Send(MCP_RESET);		   //write reset instruction
  SetSS(1);				   //unselect the slave
  delay(10);

}

extern unsigned char mcp2515_READ_Byte(unsigned char address)
{
  unsigned char res;
  SetSS(0);
  Send(0x03);
  Send(address);
  res = Recive();
 
  SetSS(1);
  delay(10);
  return res;
}

extern unsigned char mcp2515_WRITE_Byte(unsigned char address,unsigned char data)
{
  unsigned char dummy;
  SetSS(0);
  dummy =Send(0x02);
  dummy =Send(address);
  dummy =Send(data);
  SetSS(1);
  delay(10);
  return dummy;
}

extern mcp2515_READ(unsigned char address,unsigned char *data,unsigned int size)
{
  SetSS(0);
  Send(0x03);
  Send(address);
  RecBuf(data,size);  
  SetSS(1);
  delay(10);
  
}

extern mcp2515_WRITE(unsigned char address,unsigned char *data,unsigned int size)
{
  
  SetSS(0);
  Send(MCP_WRITE);
  Send(address);
   SendBuf(data,size);
  SetSS(1);
  delay(10);
  
}

extern mcp2515_modify(unsigned char address,unsigned char mask,unsigned char data)
{
	
	SetSS(0);
	Send(MCP_BITMOD);
	Send(address);
	Send(mask);
	Send(data);
	SetSS(1);

}

extern unsigned char mcp2515_readXXStatus_helper(unsigned char cmd)
{
	unsigned char i;
	
	SetSS(0);
	i = Send(cmd);
	SetSS(1);
	
	return i;
}

extern unsigned char mcp2515_readStatus(void)
{
	return mcp2515_readXXStatus_helper(MCP_READ_STATUS);
}

extern unsigned char mcp2515_RXStatus(void)
{
	return mcp2515_readXXStatus_helper(MCP_RX_STATUS);
}

/************** CAN CONTROL FUNCTIONS USE MCP2515 *********************/

extern unsigned char mcp2515_setCANCTRL_Mode(unsigned char newmode)
{
	unsigned char i;
	
	mcp2515_modify(MCP_CANCTRL, MODE_MASK, newmode);
	
	// verify as advised in datasheet
	i = mcp2515_READ_Byte(MCP_CANCTRL);
	i &= MODE_MASK;
	
	if ( i == newmode ) {
		return MCP2515_OK; 
	}
	else {
		return MCP2515_FAIL;
	}
}

extern unsigned char mcp2515_configRate(unsigned char canSpeed)
{
	unsigned char set, cfg1, cfg2, cfg3;
	
	set = 0;
	
	switch (canSpeed) {
		case (CAN_125KBPS) :
			cfg1 = MCP_4MHz_125kBPS_CFG1 ;
			cfg2 = MCP_4MHz_125kBPS_CFG2 ;
			cfg3 = MCP_4MHz_125kBPS_CFG3 ;
			set = 1;
			break;
		case (CAN_20KBPS) :
			cfg1 = MCP_4MHz_20kBPS_CFG1 ;
			cfg2 = MCP_4MHz_20kBPS_CFG2 ;
			cfg3 = MCP_4MHz_20kBPS_CFG3 ;
			set = 1;
			break;
		default:
			set = 0;
			break;
	}
	
	if (set) {
		mcp2515_WRITE_Byte(MCP_CNF1, cfg1);
		mcp2515_WRITE_Byte(MCP_CNF2, cfg2);
		mcp2515_WRITE_Byte(MCP_CNF3, cfg3);
		return MCP2515_OK;
	}
	else {
		return MCP2515_FAIL;
	}
} 

extern mcp2515_read_can_id( unsigned char mcp_addr,unsigned char *ext, unsigned int *can_id )
{
    unsigned char tbufdata[4];
	
    *ext = 0;
    *can_id = 0;
    
	mcp2515_READ( mcp_addr, tbufdata, 4 );
    
	*can_id = (tbufdata[MCP_SIDH]<<3) + (tbufdata[MCP_SIDL]>>5);
	
    if ( (tbufdata[MCP_SIDL] & MCP_TXB_EXIDE_M) ==  MCP_TXB_EXIDE_M ) {
		// extended id
        *can_id = (*can_id<<2) + (tbufdata[MCP_SIDL] & 0x03);
        *can_id <<= 16;
        *can_id = *can_id +(tbufdata[MCP_EID8]<<8) + tbufdata[MCP_EID0];
        *ext = 1;
    }
}

// Buffer can be MCP_RXBUF_0 or MCP_RXBUF_1

extern mcp2515_read_canMsg( unsigned char buffer_sidh_addr,CanMessage* msg)
{

    unsigned char mcp_addr, ctrl;

	mcp_addr = buffer_sidh_addr;
	
    mcp2515_read_can_id( mcp_addr, &(msg->extended_identifier), 
		&(msg->identifier) );
    
	ctrl = mcp2515_READ_Byte( mcp_addr-1 );
    msg->dlc = mcp2515_READ_Byte( mcp_addr+4 );
    
	if (/*(*dlc & RTR_MASK) || */(ctrl & 0x08)) {
        msg->rtr = 1;
    } else {
        msg->rtr = 0;
    }
    
	msg->dlc &= MCP_DLC_MASK;
    mcp2515_READ( mcp_addr+5, &(msg->dta[0]), msg->dlc );
}

extern mcp2515_write_can_id( unsigned char mcp_addr,unsigned char ext, unsigned int can_id )
{
    unsigned int canid;
    unsigned char tbufdata[4];
	
    canid = (unsigned int)(can_id & 0x0FFFF);
    
	if ( ext == 1) {
        tbufdata[MCP_EID0] = (unsigned char) (canid & 0xFF);
        tbufdata[MCP_EID8] = (unsigned char) (canid / 256);
        canid = (unsigned int)( can_id / 0x10000L );
        tbufdata[MCP_SIDL] = (unsigned char) (canid & 0x03);
        tbufdata[MCP_SIDL] += (unsigned char) ((canid & 0x1C )*8);
        tbufdata[MCP_SIDL] |= MCP_TXB_EXIDE_M;
        tbufdata[MCP_SIDH] = (unsigned char) (canid / 32 );
    }
    else {
        tbufdata[MCP_SIDH] = (unsigned char) (canid / 8 );
        tbufdata[MCP_SIDL] = (unsigned char) ((canid & 0x07 )*32);
        tbufdata[MCP_EID0] = 0;
        tbufdata[MCP_EID8] = 0;
    }
	mcp2515_WRITE( mcp_addr, tbufdata, 4 );
}

// Buffer can be MCP_TXBUF_0 MCP_TXBUF_1 or MCP_TXBUF_2

extern mcp2515_write_canMsg( unsigned char buffer_sidh_addr, CanMessage* msg)
{
    unsigned char mcp_addr, dlc;

	mcp_addr = buffer_sidh_addr;
	dlc = msg->dlc;
	
    mcp2515_WRITE(mcp_addr+5, &(msg->dta[0]), dlc );  // write data bytes
    mcp2515_write_can_id( mcp_addr, msg->extended_identifier,
		msg->identifier );  // write CAN id
    if ( msg->rtr == 1)  dlc |= MCP_RTR_MASK;  // if RTR set bit in byte
    mcp2515_WRITE_Byte( (mcp_addr+4), dlc );  // write the RTR and DLC
}

/*
 ** Start the transmission from one of the tx buffers.
 */

// Buffer can be MCP_TXBUF_0 MCP_TXBUF_1 or MCP_TXBUF_2

extern mcp2515_start_transmit(unsigned char buffer_sidh_addr)
{
	// TXBnCTRL_addr = TXBnSIDH_addr - 1
  
    mcp2515_modify( buffer_sidh_addr-1 , MCP_TXB_TXREQ_M, 
		MCP_TXB_TXREQ_M );
}

extern unsigned char mcp2515_getNextFreeTXBuf(unsigned char *txbuf_n)
{
	unsigned char res, i, ctrlval;
	unsigned char ctrlregs[MCP_N_TXBUFFERS] = { MCP_TXB0CTRL, MCP_TXB1CTRL, MCP_TXB2CTRL };
	
	res = MCP_ALLTXBUSY;
	*txbuf_n = 0x00;
	
	// check all 3 TX-Buffers
	for (i=0; i<MCP_N_TXBUFFERS; i++) {
		ctrlval = mcp2515_READ_Byte( ctrlregs[i] );
		if ( (ctrlval & MCP_TXB_TXREQ_M) == 0 )
		 {
#if (MCPDEBUG_TXBUF)
			Debug_ByteToUart_P("Selected TX-Buffer", i+1);
#endif
			*txbuf_n = ctrlregs[i]+1; // return SIDH-address of Buffer
			res = MCP2515_OK;
			return res; /* ! function exit */
		}
	}
	
	return res;
}

extern mcp2515_initCANBuffers(void)
{
	unsigned char i, a1, a2, a3;
	
	// TODO: check why this is needed to receive extended 
	//   and standard frames
	// Mark all filter bits as don't care:
    mcp2515_write_can_id(MCP_RXM0SIDH, 0, 0);
    mcp2515_write_can_id(MCP_RXM1SIDH, 0, 0);
    // Anyway, set all filters to 0:
    mcp2515_write_can_id(MCP_RXF0SIDH, 1, 0); // RXB0: extended 
    mcp2515_write_can_id(MCP_RXF1SIDH, 0, 0); //       AND standard
    mcp2515_write_can_id(MCP_RXF2SIDH, 1, 0); // RXB1: extended 
    mcp2515_write_can_id(MCP_RXF3SIDH, 0, 0); //       AND standard
    mcp2515_write_can_id(MCP_RXF4SIDH, 0, 0);
    mcp2515_write_can_id(MCP_RXF5SIDH, 0, 0);
	
	// Clear, deactivate the three transmit buffers
	// TXBnCTRL -> TXBnD7
    a1 = MCP_TXB0CTRL;
	a2 = MCP_TXB1CTRL;
	a3 = MCP_TXB2CTRL;
    for (i = 0; i < 14; i++) { // in-buffer loop
		mcp2515_WRITE_Byte(a1, 0);
		mcp2515_WRITE_Byte(a2, 0);
		mcp2515_WRITE_Byte(a3, 0);
        a1++;
		a2++;
		a3++;
    }
	
    // and clear, deactivate the two receive buffers.
    mcp2515_WRITE_Byte(MCP_RXB0CTRL, 0);
    mcp2515_WRITE_Byte(MCP_RXB1CTRL, 0);
}

extern unsigned char mcp2515_init(unsigned char canSpeed)
{
	unsigned char res;
	
	SetSS(1);
	//MCP_CS_DDR |= ( 1 << MCP_CS_BIT );
	
	mcp2515_RESET();
	
	res = mcp2515_setCANCTRL_Mode(MODE_CONFIG);
	
	if ( res == MCP2515_FAIL ) 
	 return res;  /* function exit on error */
	res = mcp2515_configRate(canSpeed);
	

	if ( res == MCP2515_OK ) {
		mcp2515_initCANBuffers();
	}
	
	return res;
}
